#pragma once

#include <cmath>
#include <functional>
#include <string>

#include "driver/gpio.h"
#include "lwip/sockets.h"

// undefining IPADDR_NONE from lwip because it is redefined in arduino
#undef INADDR_NONE

namespace qthing {

// types
enum pull_resistor_t { none, pullup, pulldown };
enum connection_status_t { uninitialized, disconnected, connecting, connected };
using message_callback_t = std::function<void(std::string)>;
using connection_status_callback_t = std::function<void(connection_status_t)>;
using mqtt_connected_callback_t = std::function<void()>;

struct multipart_message_t {
  const std::string& topic;
  const char* payload;
  const uint32_t length;
  const uint32_t offset;
  const uint32_t total_length;
};

using binary_message_callback_t = std::function<void(multipart_message_t)>;

enum ota_state_t { start, progress, success, error };
struct ota_event_t {
  const ota_state_t state;
  const esp_err_t error;
  const uint32_t bytes_written;
  const uint32_t bytes_total;
};

using ota_event_callback_t = std::function<void(ota_event_t)>;

// core
void enable_status_led(gpio_num_t gpio = GPIO_NUM_2, bool negated = false);
void show_activity();

// mqtt
void publish_message(const std::string& topic, const std::string& message);
void add_message_callback(const std::string& topic, message_callback_t callback);
void add_binary_message_callback(const std::string& topic, binary_message_callback_t callback);
void add_ota_event_callback(ota_event_callback_t callback);
connection_status_t get_mqtt_connection_status();
bool is_mqtt_connected();
void add_mqtt_connection_status_callback(connection_status_callback_t callback);
void add_mqtt_connected_callback(mqtt_connected_callback_t callback);

// common IO
void enable_oled();
void log_oled(const std::string& message);

void add_digital_input(gpio_num_t gpio_num, pull_resistor_t pull_resistor, std::function<void()> on_falling_edge,
                       std::function<void()> on_rising_edge);
void add_digital_input(gpio_num_t gpio_num, pull_resistor_t pull_resistor, std::string topic);

void add_button(gpio_num_t gpio, std::function<void()> on_press = NULL, std::function<void()> on_release = NULL);
void add_button(gpio_num_t gpio, std::string topic, std::string message);
void add_relay(const std::string& topic, gpio_num_t gpio_off, gpio_num_t gpio_on);
void add_digital_output(const std::string& topic, gpio_num_t gpio);

void enable_lcd(const std::string& topic, uint8_t rs, uint8_t en, uint8_t d4, uint8_t d5, uint8_t d6, uint8_t d7);

// LED strips
enum led_strip_types_t {
  LED_WS2812_V1,
  LED_WS2812B_V1,
  LED_WS2812B_V2,
  LED_WS2812B_V3,
  LED_WS2813_V1,
  LED_WS2813_V2,
  LED_WS2813_V3,
  LED_SK6812_V1,
  LED_SK6812W_V1,
};

union led_color_t {
  struct __attribute__((packed)) {
    uint8_t r, g, b, w;
  };
  uint32_t num;

  led_color_t(uint8_t r, uint8_t g, uint8_t b, uint8_t w) : r(r), g(g), b(b), w(w) {}
  led_color_t() : num(0) {}

  led_color_t operator+(const led_color_t& other) {
    led_color_t result = *this;
    result.r += other.r;
    result.g += other.g;
    result.b += other.b;
    result.w += other.w;
    return result;
  }
  led_color_t operator*(const float scale) {
    led_color_t result = *this;
    result.r *= scale;
    result.g *= scale;
    result.b *= scale;
    result.w *= scale;
    return result;
  }
};

// colors
struct RGBW {
  float r;
  float g;
  float b;
  float w;
  RGBW() : r(0), g(0), b(0), w(0) {}
  RGBW(float r, float g, float b, float w) : r(r), g(g), b(b), w(w) {}
  RGBW operator+(const RGBW& other) {
    RGBW result = *this;
    result.r += other.r;
    result.g += other.g;
    result.b += other.b;
    result.w += other.w;
    return result;
  }
  RGBW operator*(const float scale) {
    RGBW result = *this;
    result.r *= scale;
    result.g *= scale;
    result.b *= scale;
    result.w *= scale;
    return result;
  }
  void operator*=(const float scale) {
    this->r *= scale;
    this->g *= scale;
    this->b *= scale;
    this->w *= scale;
  }

  operator led_color_t() { return led_color_t(this->r * 255, this->g * 255, this->b * 255, this->w * 255); }
};

struct RGB {
  float r;
  float g;
  float b;
  RGB() : r(0), g(0), b(0) {}
  RGB(float r, float g, float b) : r(r), g(g), b(b) {}
  RGB operator+(const RGB& other) {
    RGB result = *this;
    result.r += other.r;
    result.g += other.g;
    result.b += other.b;
    return result;
  }
  RGB operator*(const float scale) {
    RGB result = *this;
    result.r *= scale;
    result.g *= scale;
    result.b *= scale;
    return result;
  }
  void operator*=(const float scale) {
    this->r *= scale;
    this->g *= scale;
    this->b *= scale;
  }

  operator RGBW() { return RGBW(this->r, this->g, this->b, 0); }

  operator led_color_t() { return led_color_t(this->r * 255, this->g * 255, this->b * 255, 0); }
};

struct GammaCorrection {
  float gamma_r;
  float gamma_g;
  float gamma_b;
  float gamma_w;
  float max_r;
  float max_g;
  float max_b;
  float max_w;
  GammaCorrection(const float gamma_r, const float gamma_g, const float gamma_b, const float gamma_w, const float max_r,
                  const float max_g, const float max_b, const float max_w)
      : gamma_r(gamma_r),
        gamma_g(gamma_g),
        gamma_b(gamma_b),
        gamma_w(gamma_w),
        max_r(max_r),
        max_g(max_g),
        max_b(max_b),
        max_w(max_w) {}
  GammaCorrection(const float gamma_r, const float gamma_g, const float gamma_b, const float gamma_w)
      : GammaCorrection(gamma_r, gamma_g, gamma_b, gamma_w, 1, 1, 1, 1) {}
  GammaCorrection(const float gamma) : GammaCorrection(gamma, gamma, gamma, gamma) {}
  const RGBW operator()(const RGBW& color) {
    return RGBW(pow(color.r, gamma_r) * max_r, pow(color.g, gamma_g) * max_g, pow(color.b, gamma_b) * max_b,
                pow(color.w, gamma_w) * max_w);
  }
};

RGB hue(const float hue, const float value = 1);
RGB hsv(float hue, float staturation = 1, float value = 1);

using led_color_handler_t = std::function<led_color_t(uint8_t, uint16_t, led_color_t)>;
using led_setup_handler_t = std::function<void()>;

void enable_leds(uint8_t led_strip_fps = 60);
void add_ledstrip(gpio_num_t gpio, uint16_t num_leds, qthing::led_strip_types_t led_type);
led_color_handler_t get_color_handler();
void set_color_handler(led_color_handler_t led_color_handler);
void set_led_strip_fps(uint8_t fps);

void configure_network_color_handler();

using ntp_callback_t = std::function<void()>;
void enable_ntp(std::string address = "pool.ntp.org", std::string timezone = "CET-1CEST,M3.5.0,M10.5.0/3",
                qthing::ntp_callback_t callback = NULL);

typedef struct {
  struct sockaddr_in6 sourceAddr;
  const char* payload;
  const uint16_t length;
} udpPacket;

using udpPacketCallback = std::function<void(udpPacket)>;
void addUDPPacketCallback(std::string magicString, udpPacketCallback callback);
void start_udp_server();

// peripherals
namespace button_panel_5x5 {
void initialize(std::function<void(uint8_t, uint8_t, bool)> callback = NULL);
bool get_key_state(uint8_t x, uint8_t y);
uint32_t get_key_activation_time(uint8_t x, uint8_t y);
}  // namespace button_panel_5x5

// utility
std::string to_string(float value, int precision = 2);
std::string to_string(double value, int precision = 2);
std::string to_string(uint32_t value);
std::string to_string(struct sockaddr_in6 sourceAddr);

}  // namespace qthing
