#include <time.h>

#include "Arduino.h"
#include "SSD1306Wire.h"
#include "qthing/event.hpp"
#include "qthing/network.hpp"
#include "qthing/oled_symbols.hpp"
#include "qthing/qthing_legacy.hpp"

namespace qthing {

// SSD1306Wire display(0x3c, GPIO_NUM_5, GPIO_NUM_4);
SSD1306Wire display = SSD1306Wire(0x3c, GPIO_NUM_4, GPIO_NUM_15);

#define OLED_LOG_MAX_LENGTH 5
std::string oled_log_messages[OLED_LOG_MAX_LENGTH];
uint8_t oled_log_first_message = 0;
uint8_t oled_log_message_count = 0;

volatile long redraw_time = 1;

#define FRAME_TIME_MS 100

#define RESET_AFTER_MS 10000
long last_reset = 0;

void redraw() { redraw_time = millis(); }

void redraw_at_time(long time_in_ms) {
  if (redraw_time == 0 || time_in_ms < redraw_time) {
    redraw_time = time_in_ms;
  }
}

void redraw_after_timespan(long timespan_in_ms) { redraw_at_time(millis() + timespan_in_ms); }

void draw_network_status(uint16_t& x, uint16_t y) {
  x -= 10;
  connection_status_t wifi_status = get_wifi_connection_status();
  if (wifi_status == connected) {
    display.drawFastImage(x, y, 10, 8, wifi_connected_symbol);
  } else if (wifi_status == connecting) {
#define ICON_UPDATE_RATE 200
    int m = millis() / ICON_UPDATE_RATE % 3;
    if (m == 0) {
      display.drawFastImage(x, y, 10, 8, wifi_connecting_symbol_1);
    } else if (m == 1) {
      display.drawFastImage(x, y, 10, 8, wifi_connecting_symbol_2);
    } else {
      display.drawFastImage(x, y, 10, 8, wifi_connecting_symbol_3);
    }
    redraw_after_timespan(ICON_UPDATE_RATE);
  } else if (wifi_status == disconnected) {
    display.drawFastImage(x, y, 10, 8, wifi_disconnected_symbol);
  }
}

void draw_mqtt_status(uint16_t& x, uint16_t y) {
  x -= 8;
  connection_status_t mqtt_status = get_mqtt_connection_status();
  if (mqtt_status == connected) {
    display.drawFastImage(x, y, 8, 8, mqtt_connected_symbol);
  } else if (mqtt_status == connecting) {
    display.drawFastImage(x, y, 8, 8, mqtt_connecting_symbol);
  } else if (mqtt_status == disconnected) {
    display.drawFastImage(x, y, 8, 8, mqtt_disconnected_symbol);
  }
}

void draw_text_indicator(uint16_t& x, uint16_t y, String text) {
  int width = display.getStringWidth(text);
  x -= width;
  // offset of -2 to align with symbols (font has height 10, symbols 8)
  display.drawString(x, y - 2, text);
}

bool draw_clock(uint16_t& x, uint16_t y) {
  time_t t = time(0);
  tm* now = localtime(&t);
  if (now->tm_year < 100) {
    return false;
  }

  auto hour = now->tm_hour;
  auto minute = now->tm_min;

  char time[6];
  time[0] = '0' + hour / 10;
  time[1] = '0' + hour % 10;
  time[2] = ':';
  time[3] = '0' + minute / 10;
  time[4] = '0' + minute % 10;
  time[5] = 0;

  draw_text_indicator(x, y, time);
  redraw_after_timespan((60 - now->tm_sec) * 1000);

  return true;
}

void display_status() {
  uint16_t x = 128;
  const uint16_t y = 2;

  bool clock = draw_clock(x, y);
  if (clock) {
    x -= 2;
  }

  draw_network_status(x, y);
  x -= 2;
  draw_mqtt_status(x, y);
  x -= 2;
}

void display_log() {
  // display.setFont(ArialMT_Plain_10);
  // display.setTextAlignment(TEXT_ALIGN_LEFT);
  for (uint8_t i = 0; i < oled_log_message_count; i += 1) {
    uint8_t line = (i + oled_log_first_message) % OLED_LOG_MAX_LENGTH;
    display.drawString(0, 10 + (i * 10), oled_log_messages[line].c_str());
  }
}

void oled_task(void* pvParameters) {
  while (true) {
    long time_ms = millis();
    bool reset = last_reset + RESET_AFTER_MS < time_ms;

    if (reset || (redraw_time > 0 && redraw_time <= time_ms)) {
      redraw_time = 0;

      if (reset) {
        display.resetDisplay();
        last_reset = time_ms;
      } else {
        display.clear();
      }

      display_log();
      display_status();

      // display.setFont(ArialMT_Plain_10);
      // display.setTextAlignment(TEXT_ALIGN_RIGHT);
      // display.drawString(64, 0, String(millis()));

      display.display();
    }

    vTaskDelay(pdMS_TO_TICKS(FRAME_TIME_MS));
  }
}

void oled_log(const std::string& message) {
  if (oled_log_message_count < OLED_LOG_MAX_LENGTH) {
    oled_log_messages[oled_log_message_count] = message;
    oled_log_message_count += 1;
  } else {
    oled_log_messages[oled_log_first_message] = message;
    oled_log_first_message += 1;
    oled_log_first_message %= OLED_LOG_MAX_LENGTH;
  }
  redraw();
}

void enable_oled() {
  oled_log("qthing");
  oled_log("much wow, very secure");

  // display = SSD1306Wire(0x3c, GPIO_NUM_4, GPIO_NUM_15);

  set_oled_log_callback(oled_log);
  display.init();
  display.flipScreenVertically();
  display.setFont(ArialMT_Plain_10);
  xTaskCreate(&oled_task, "oled_task", 4096, NULL, 10, NULL);
}

}  // namespace qthing
