#include <driver/gpio.h>
#include <esp_log.h>
#include <freertos/FreeRTOS.h>
#include <freertos/task.h>

#include <array>
#include <functional>

#include "qthing/io.hpp"
#include "qthing/qthing_legacy.hpp"

#define TAG "button_panel"

#define WIDTH 5
#define HEIGHT 5
std::array<bool, WIDTH * HEIGHT> key_states;
std::array<uint32_t, WIDTH * HEIGHT> activation_times;

std::array<gpio_num_t, WIDTH> inputs{GPIO_NUM_32, GPIO_NUM_35, GPIO_NUM_34, GPIO_NUM_39, GPIO_NUM_36};
std::array<gpio_num_t, HEIGHT> outputs{GPIO_NUM_33, GPIO_NUM_12, GPIO_NUM_14, GPIO_NUM_4, GPIO_NUM_15};

std::function<void(uint8_t, uint8_t, bool)> key_callback;

void read_button_panel() {
  for (uint8_t y = 0; y < HEIGHT; y += 1) {
    gpio_num_t output = outputs[y];
    ESP_ERROR_CHECK(gpio_set_level(output, 1));
    // sleep for one port tick to ensure level is set
    vTaskDelay(1);
    // ESP_LOGI(TAG, "Reading row y=%i", y);
    for (uint8_t x = 0; x < WIDTH; x += 1) {
      gpio_num_t input = inputs[x];
      uint8_t address = x + y * WIDTH;
      bool last_state = key_states[address];
      bool state = gpio_get_level(input);
      if (state != last_state) {
        ESP_LOGI(TAG, "Key state changed: (%i, %i) is %i", x, y, state);
        key_states[address] = state;
        activation_times[address] = xTaskGetTickCount() * portTICK_PERIOD_MS;
        if (key_callback != NULL) {
          key_callback(x, y, state);
        }
      }
    }
    ESP_ERROR_CHECK(gpio_set_level(output, 0));
  }
}

void button_panel_task(void* _) {
  while (true) {
    read_button_panel();
  }
}

namespace qthing {
namespace button_panel_5x5 {
void initialize(std::function<void(uint8_t, uint8_t, bool)> callback) {
  key_callback = callback;

  for (gpio_num_t input : inputs) {
    ESP_LOGI(TAG, "Initializing gpio %i as input", input);
    gpio_init_input(input);
  }

  for (gpio_num_t output : outputs) {
    ESP_LOGI(TAG, "Initializing gpio %i as output", output);
    gpio_pad_select_gpio(output);
    ESP_ERROR_CHECK(gpio_set_direction(output, GPIO_MODE_OUTPUT));
    ESP_ERROR_CHECK(gpio_set_level(output, 1));
  }

  xTaskCreate(button_panel_task, "button_panel_task", 4096, NULL, 6, NULL);
}

bool get_key_state(uint8_t x, uint8_t y) { return key_states[x + y * WIDTH]; }

uint32_t get_key_activation_time(uint8_t x, uint8_t y) { return activation_times[x + y * WIDTH]; }
}  // namespace button_panel_5x5
}  // namespace qthing
